//===================================================================
// Basic template, routines and definitions for the                 =
// "Ant51"-Board 1V1. Tested with SDCC V2.3.0                      =
//                                                                  =
// Used MCU : Atmel AT89S53 ( 256 Byte RAM, 12 KByte FLASH,         =
//                            3 Timer, UART, @ 24 MHz )             =
//                                                                  =
// Compiler commandline :                                           =
//                                                                  =
//              sdcc uAnt51T0.c --stack-after-data                  =
//                                                                  =
// Identifier : bit    -> located in bit adressable RAM             =
//              idata  -> located in upper 128 byte of RAM          =
//                        (available only at 256 byte RAM C)       =
//              code   -> located in PROM/EPROM/FLASH               =
//                                                                  =
// ToDos      : - I2C-Timeout                                       =
//              - SetBit/GetBit for XIO                             =
//              - SetAlarm for RTC                                  =
//              - RAM_Read/RAM_Write for unused RTC-RAM             =
//                                                                  =
//   (C) 2001/2002 by Dipl.-Ing.(FH) Thorsten Godau DL9SEC          =
//===================================================================

//===================================================================
// Port-map of the MCU :                                            =
//                                                                  =
// P0_0 -> unused                                                   =
// P0_1 -> unused                                                   =
// P0_2 -> unused                                                   =
// P0_3 -> unused                                                   =
// P0_4 -> unused                                                   =
// P0_5 -> unused                                                   =
// P0_6 -> unused                                                   =
// P0_7 -> unused                                                   =
//                                                                  =
// P1_0 -> T2 / I2C SDA                                             =
// P1_1 -> T2EX / I2C SCL                                           =
// P1_2 -> BEEPER                                                   =
// P1_3 -> LED1                                                     =
// P1_4 -> #SS (internal, SPI-Slave-Select)                         =
// P1_5 -> SPI MOSI                                                 =
// P1_6 -> SPI MISO                                                 =
// P1_7 -> SPI SCK                                                  =
//                                                                  =
// P2_0 -> unused                                                   =
// P2_1 -> unused                                                   =
// P2_2 -> LCD RS                                                   =
// P2_3 -> LCD E                                                    =
// P2_4 -> LCD D4                                                   =
// P2_5 -> LCD D5                                                   =
// P2_6 -> LCD D6                                                   =
// P2_7 -> LCD D7                                                   =
//                                                                  =
// P3_0 -> V.24 Rx-Data                                             =
// P3_1 -> V.24 Tx-Data                                             =
// P3_2 -> INT0 / INTXIO (PCF8574)                                  =
// P3_3 -> INT1 / INTRTC (PCF8583)                                  =
// P3_4 -> T0 / LCD Backlight / LED4                                =
// P3_5 -> T1 / LCD R/W                                             =
// P3_6 -> [WR#] unused                                             =
// P3_7 -> [RD#] unused                                             =
//                                                                  =
//===================================================================

//===================================================================
// Memory-map of the EEPROM (M24C02, 256 Byte) :                    =
//                                                                  =
// 0x00 ->                                                          =
// 0x01 ->                                                          =
//===================================================================

//===================================================================
// Connections to the ADC (PCF8591) :                               =
//                                                                  =
// Ch 0 ->                                                          =
// Ch 1 ->                                                          =
// Ch 2 ->                                                          =
// Ch 4 ->                                                          =
//                                                                  =
// DAC  ->                                                          =
//===================================================================

//===================================================================
// Connections to the DPOT (DS1803) :                               =
//                                                                  =
// Pot 0 ->                                                         =
// Pot 1 ->                                                         =
//===================================================================

//===================================================================
// Errorcodes generated by errormsg() :                             =
//                                                                  =
// '0' -> unused                                                   =
// '1' -> EEPROM_write : SLAVE adress write failure                =
// '2' -> EEPROM_write : memory adress write failure               =
// '3' -> EEPROM_write : memory data write failure                 =
// '4' -> EEPROM_read  : SLAVE adress dummy write failure          =
// '5' -> EEPROM_read  : memory adress dummy write failure         =
// '6' -> EEPROM_read  : SLAVE adress write failure                =
// '7' -> DPOT_write   : SLAVE adress write failure                =
// '8' -> DPOT_write   : command write failure                     =
// '9' -> DPOT_write   : data write failure                        =
// 'A' -> ADC_read     : SLAVE (W-Flag) adress write failure       =
// 'B' -> ADC_read     : control byte write failure                =
// 'C' -> ADC_read     : SLAVE (R-Flag) adress write failure       =
// 'D' -> DAC_write    : SLAVE (W-Flag) adress write failure       =
// 'E' -> DAC_write    : control byte write failure                =
// 'F' -> DAC_write    : content byte write failure                =
// 'G' -> RTC write    : SLAVE (W-Flag) adress write failure       =
// 'H' -> RTC write    : memory adress write failure               =
// 'I' -> RTC write    : SLAVE (R-Flag) adress write failure       =
// 'J' -> RTC write    : Parameter write failure                   =
// 'K' -> XIO_write    : SLAVE (W-Flag) adress write failure       =
// 'L' -> XIO_write    : data write failure                        =
// 'M' -> RTC write    : SLAVE (R-Flag) adress write failure       =
//===================================================================


// INCLUDES =========================================================
#include "inc/AT89S53.h"      // Definitions of registers, SFRs and Bits
#include "inc/uAnt51.h"       // Global definitions and declarations

// END INCLUDES =====================================================


// ISR-Prototypes ===================================================
void External0_ISR(void) interrupt 0; // ISR for the external input INT0
void Timer0_ISR(void)    interrupt 1; // ISR for Timer0 overflow
void External1_ISR(void) interrupt 2; // ISR for the external input INT1
void Timer1_ISR(void)    interrupt 3; // ISR for Timer1 overflow
void Serial_ISR(void)    interrupt 4; // ISR for serial reception
void Timer2_ISR(void)    interrupt 5; // ISR for Timer2 overflow

// END ISR-Prototypes ===============================================


// Prototypes =======================================================
void softbacklight(bit direction);      // Smooth turn on/off the LCD backlight

void errormsg(unsigned char ascii);     // Generates a error message via UART
void init_MCU(void);                    // Initialize the MCU
void delay(unsigned int time_100us);    // Timer dependend delay routine
void putchar(unsigned char value);      // Sends a character via UART
void UART_puts(unsigned char *text);    // Sends a string to the UART
unsigned int RXChar(void);              // Reads a character out of the RX buffer

void LCD_init(void);                    // Initialize the LCD display
void LCD_clrscr(void);                  // Clears LCD screen
void LCD_putchar(unsigned char value);  // Writes a character to display
// Prints a text to x/y position
void LCD_printxy(unsigned char x,unsigned char y,unsigned char *text);
// Controls the display
void LCD_control(unsigned char dsp,unsigned char blink,unsigned char cursor);
// Sets LCD write position
void LCD_gotoxy(unsigned char x,unsigned char y);
// Download custom characters to the LCD
void LCD_customcharset(unsigned char *pattern);

void i2c_delay(void);                         // Wait for 5s
void i2c_start(void);                         // Geneate I2C START condition
void i2c_stop(void);                          // Geneate I2C STOP condition
void i2c_ack(void);                           // Geneate I2C MASTER ACK
void i2c_nack(void);                          // Geneate I2C MASTER NACK
unsigned char i2c_write(unsigned char value); // Write a character via I2C
unsigned char i2c_check(void);                // Check an ACK from SLAVE
unsigned char i2c_read(void);                 // Read a character via I2C

// Read byte from EEPROM
unsigned char EEPROM_read(unsigned char adress);
// Write byte to EEPROM
void EEPROM_write(unsigned char adress,unsigned char content);

// Write byte to DPOT
void DPOT_write(bit potnum,unsigned char content);

// Read byte of ADC[ch]
unsigned char ADC_read(unsigned char adchannel);

// Write byte to DAC
void DAC_write(unsigned char content);

// Get time from RTC (result is in TM)
void gettime(struct TIME_T *time);

// Set time of RTC (TM to be filled)
void settime(struct TIME_T *time);

// Write byte to XIO
void XIO_write(unsigned char content);

// Read byte from XIO
unsigned char XIO_read(void);


// END Prototypes ===================================================


/********************************************************************/
/* Function    : External0_ISR()                                ISR */
/*------------------------------------------------------------------*/
/* Description : Interrupt-Service-Routine on vector #0 for         */
/*               handling interrupt requests on external pin P3_2.  */
/*               Free for further use.                              */
/*------------------------------------------------------------------*/
/* Author      : Dipl.-Ing. (FH) Thorsten Godau                     */
/*------------------------------------------------------------------*/
/* Input       : none                                               */
/*------------------------------------------------------------------*/
/* Returnvalue : none                                               */
/*------------------------------------------------------------------*/
/* History     : 06/99    V1.0 Basic routine                        */
/*                                                                  */
/********************************************************************/
void External0_ISR(void) interrupt 0
{

EX0 = 0;

// Safety-routine
_asm
  nop
_endasm;

//EX0 = 1;

return;
}


/********************************************************************/
/* Function    : Timer0_ISR()                                   ISR */
/*------------------------------------------------------------------*/
/* Description : Interrupt-Service-Routine on vector #1 for         */
/*               handling Timer0 interrupts.                        */
/*               Free for further use.                              */
/*------------------------------------------------------------------*/
/* Author      : Dipl.-Ing. (FH) Thorsten Godau                     */
/*------------------------------------------------------------------*/
/* Input       : none                                               */
/*------------------------------------------------------------------*/
/* Returnvalue : none                                               */
/*------------------------------------------------------------------*/
/* History     : 10/01    V1.0 Basic routine                        */
/*                                                                  */
/********************************************************************/
void Timer0_ISR(void) interrupt 1
{

ET0 = 0;                     // Disable Timer0-interrupt

// Safety-routine
_asm
  nop
_endasm;

//ET0 = 1;                     // Enable Timer0-interrupt

return;
}


/********************************************************************/
/* Function    : External1_ISR()                                ISR */
/*------------------------------------------------------------------*/
/* Description : Interrupt-Service-Routine on vector #2 for         */
/*               handling interrupt requests on external pin P3_3.  */
/*               Free for further use.                              */
/*------------------------------------------------------------------*/
/* Author      : Dipl.-Ing. (FH) Thorsten Godau                     */
/*------------------------------------------------------------------*/
/* Input       : none                                               */
/*------------------------------------------------------------------*/
/* Returnvalue : none                                               */
/*------------------------------------------------------------------*/
/* History     : 06/99    V1.0 Basic routine                        */
/*                                                                  */
/********************************************************************/
void External1_ISR(void) interrupt 2
{

EX1 = 0;

// Safety-routine
_asm
  nop
_endasm;

//EX1 = 1;

return;
}


/********************************************************************/
/* Function    : Timer1_ISR()                                   ISR */
/*------------------------------------------------------------------*/
/* Description : Interrupt-Service-Routine on vector #3 for         */
/*               handling Timer1 interrupts.                        */
/*               Timer1 ticks every 100s and is used as global     */
/*               delay-timer in delay().                            */
/*------------------------------------------------------------------*/
/* Author      : Dipl.-Ing. (FH) Thorsten Godau                     */
/*------------------------------------------------------------------*/
/* Input       : none                                               */
/*------------------------------------------------------------------*/
/* Returnvalue : none                                               */
/*------------------------------------------------------------------*/
/* History     : 10/01    V1.0 Basic routine                        */
/*                                                                  */
/********************************************************************/
void Timer1_ISR(void) interrupt 3
{

ET1 = 0;

if ( DELAY > 0 ) DELAY--;    // Timebase ticks for delay-routine

ET1 = 1;

return;
}


/********************************************************************/
/* Function    : Serial_ISR()                                   ISR */
/*------------------------------------------------------------------*/
/* Description : Interrupt-Service-Routine on vector #4 for         */
/*               handling serial interrupts.                        */
/*               Both character rx and tx is interrupt controlled   */
/*               and handled through ring buffers of size BSIZE.    */
/*------------------------------------------------------------------*/
/* Author      : Dipl.-Ing. (FH) Thorsten Godau                     */
/*------------------------------------------------------------------*/
/* Input       : none                                               */
/*------------------------------------------------------------------*/
/* Returnvalue : none                                               */
/*------------------------------------------------------------------*/
/* History     : 06/99    V1.0 Basic routine                        */
/*                                                                  */
/********************************************************************/
void Serial_ISR(void) interrupt 4
{
ES = 0;

if (RI)
  {
  RI = 0;                     // Clear Flag
  RXBuffer[RXB] = SBUF;       // Write RX data to buffer

  if ( ++RXB >= BSIZE )       // Check end of buffer
    RXB = 0;                  // Rewind
  }

if (TI)
  {
  TI = 0;
  BUSY = TXCNT;
  if (BUSY)
    {
    TXCNT--;
    SBUF = TXBuffer[TXPOS++];
    if ( TXPOS >= BSIZE ) TXPOS = 0;
    }
  }

ES = 1;

return;
}


/********************************************************************/
/* Function    : Timer2_ISR()                                   ISR */
/*------------------------------------------------------------------*/
/* Description : Interrupt-Service-Routine on vector #5 for         */
/*               handling Timer2 interrupts.                        */
/*               Routine is not used, because Timer2 generates      */
/*               the baudrate for the serial communication.         */
/*------------------------------------------------------------------*/
/* Author      : Dipl.-Ing. (FH) Thorsten Godau                     */
/*------------------------------------------------------------------*/
/* Input       : none                                               */
/*------------------------------------------------------------------*/
/* Returnvalue : none                                               */
/*------------------------------------------------------------------*/
/* History     : 10/01    V1.0 Basic routine                        */
/*                                                                  */
/********************************************************************/
void Timer2_ISR(void) interrupt 5
{

ET2 = 0;

// Safety-routine
_asm
  nop
_endasm;

//ET2 = 1;

return;
}


/********************************************************************/
/* Function    : init_MCU()                                     SUB */
/*------------------------------------------------------------------*/
/* Description : Routine initializes the used MCU registers,        */
/*               the timers and the global variables.               */
/*               Baudrate calculation :                             */
/*               RCAP2 = 65536 - ( Quarz / ( 32 * Baud ) )          */
/*               Timer 0/1 autoreload-register THx calculation :    */
/*               THx = Timebase * Quarz / 12                        */
/*------------------------------------------------------------------*/
/* Author      : Dipl.-Ing. (FH) Thorsten Godau                     */
/*------------------------------------------------------------------*/
/* Input       : none                                               */
/*------------------------------------------------------------------*/
/* Returnvalue : none                                               */
/*------------------------------------------------------------------*/
/* History     : 10/01    V1.0 Basic routine                        */
/*                                                                  */
/********************************************************************/
void init_MCU(void)
{

EA   = 0;     // Disable all interrupts
ET0  = 0;     // Disable Timer 0 interrupt
ET1  = 0;     // Disable Timer 1 interrupt
ET2  = 0;     // Disable Timer 2 interrupt
EX0  = 0;     // Disable External 0 interrupt
EX1  = 0;     // Disable External 1 interrupt
ES   = 0;     // Disable Serial interrupt
TR0  = 0;     // Stop Timer 0
TR1  = 0;     // Stop Timer 1
TR2  = 0;     // Stop Baudrate-Generator Timer 2

// Set all pins to input-mode
P0   = 0xFF;
P1   = 0xFF;
P2   = 0xFF;
P3   = 0xFF;

// Timer 0 & 1
TL0  = 0x00;    // MUST be initialized for the firt time with 0
TH0  = 0x64;    // Tick every 50s at 24MHz (unused)
TL1  = 0x00;    // MUST be initialized for the firt time with 0
TH1  = 0xC8;    // Tick every 100s at 24MHz
TMOD = 0x22;    // Autoreloadmode for both Timer 0 & 1

// Timer 2 (used for baudrate-generation)
C_T2 = 0;       // Timer-Mode
RCLK = 1;       // |
TCLK = 1;       // Set mode to Baudrate-Generator

SM1  = 1;       // |
SM0  = 0;       // Set UART to 8N1
REN  = 1;       // Enable UART reception
RCAP2H = 0xFF;  // |
RCAP2L = 0xD9;  // Set Baudrate to 19200Bd (19231Bd) at fq = 24MHz
                // 9600Bd (9615Bd) -> RCAP2 = 0xFFB2 at fq = 24MHz
RI  = 0;        // Clear Receive-Interrupt-Flag
TI  = 0;        // Clear Transmit-Interrupt-Flag

// Initialize globals
DELAY    = 0;
RXB      = 0;
RXE      = 0;
TXCNT    = 0;
TXPOS    = 0;
BUSY     = 0;

TM.tm_sec  = 0;
TM.tm_min  = 0;
TM.tm_hour = 0;
TM.tm_dom  = 0;
TM.tm_mon  = 0;
TM.tm_year = 0;
TM.tm_dow  = 0;


TR0  = 0;       // Stop Timer 0 (unused)
TR1  = 1;       // Start Timer 1
TR2  = 1;       // Start Baudrate-Generator Timer 2
ET0  = 0;       // Disable Timer 0 interrupt
ET1  = 1;       // Enable Timer 1 interrupt
ET2  = 0;       // Disable Timer 2 interrupt
ES   = 1;       // Enable Serial interrupt
EA   = 1;       // Enable all interrupts

return;
}


/********************************************************************/
/* Function    : LCD_init()                                     SUB */
/*------------------------------------------------------------------*/
/* Description : Routine initializes the LCD display recommended    */
/*               in the datasheet (4 Bit initialization)            */
/*------------------------------------------------------------------*/
/* Author      : Dipl.-Ing. (FH) Thorsten Godau                     */
/*------------------------------------------------------------------*/
/* Input       : none                                               */
/*------------------------------------------------------------------*/
/* Returnvalue : none                                               */
/*------------------------------------------------------------------*/
/* History     : 06/99    V1.0 Basic routine                        */
/*               09/99    V1.1 Timing correction of init            */
/*                                                                  */
/********************************************************************/
void LCD_init(void)
{

delay(200);             // Wait 20ms
LCD_E  = 0;
LCD_RS = CTRL_REG;      // Switch to inruction register

//Set LCD_DATA to high nibble of Software Reset
LCD_DATA = (LCD_DATA&0x0F)|0x30;
LCD_E = 1; LCD_E = 0;   // Write data to display
delay(50);              // Wait 5ms

LCD_E = 1; LCD_E = 0;   // Write data to display again (SW Reset)
delay(50);              // Wait 5ms

LCD_E = 1; LCD_E = 0;   // Write data to display again (SW Reset)
delay(50);              // Wait 5ms

// Set LCD_DATA to high nibble of Function Set (4Bit)
LCD_DATA = (LCD_DATA&0x0F)|0x20;
LCD_E = 1; LCD_E = 0;   // Write data to display
delay(4);               // Wait 400s

// Set LCD_DATA to high nibble of Function Set : 4 bit, 2 lines, 5*7 font
LCD_DATA = (LCD_DATA&0x0F)|0x20;
LCD_E = 1; LCD_E = 0;   // Write data to display
// Set LCD_DATA to lower nibble of Function Set : 4 bit, 2 lines, 5*7 font
LCD_DATA = (LCD_DATA&0x0F)|0x80;
LCD_E = 1; LCD_E = 0;   // Write data to display
delay(4);               // Wait 400s

// Set LCD_DATA to high nibble of Display On/Off Control : display off, cursor off, dont blink
LCD_DATA = (LCD_DATA&0x0F)|0x00;
LCD_E = 1; LCD_E = 0;   // Write data to display
// Set LCD_DATA to lower nibble of Display On/Off Control : display off, cursor off, dont blink
LCD_DATA = (LCD_DATA&0x0F)|0x80;
LCD_E = 1; LCD_E = 0;   // Write data to display
delay(4);               // Wait 400s

// Set LCD_DATA to high nibble of Clear Display
LCD_DATA = (LCD_DATA&0x0F)|0x00;
LCD_E = 1; LCD_E = 0;   // Write data to display
// Set LCD_DATA to lower nibble of Clear Display
LCD_DATA = (LCD_DATA&0x0F)|0x10;
LCD_E = 1; LCD_E = 0;   // Write data to display
delay(50);              // Wait 5ms

// Set LCD_DATA to high nibble of Entry Mode Set : increment DD-RAM address, move cursor
LCD_DATA = (LCD_DATA&0x0F)|0x00;
LCD_E = 1; LCD_E = 0;   // Write data to display
// Set LCD_DATA to lower nibble of Entry Mode Set : increment DD-RAM address, move cursor
LCD_DATA = (LCD_DATA&0x0F)|0x60;
LCD_E = 1; LCD_E = 0;   // Write data to display
delay(4);               // Wait 400s

return;
}


/********************************************************************/
/* Function    : LCD_customcharset(*pattern)                    SUB */
/*------------------------------------------------------------------*/
/* Description : Routine loads 8 customizable characters to the     */
/*               LCD. The characters are defined in CUSTOM[] and    */
/*               can be accessed as ASCII-characters 0..7           */
/*------------------------------------------------------------------*/
/* Author      : Dipl.-Ing. (FH) Thorsten Godau                     */
/*------------------------------------------------------------------*/
/* Input       : unsigned char *pattern : Pointer to character-     */
/*                                        pattern CUSTOM[64]        */
/*------------------------------------------------------------------*/
/* Returnvalue : none                                               */
/*------------------------------------------------------------------*/
/* History     : 02/02    V1.0 Basic routine                        */
/*                                                                  */
/********************************************************************/
void LCD_customcharset(unsigned char *pattern)
{
unsigned char count;    // Counter

// Set CGRAM-adress to 0x00

EA = 0;                 // Disable all interrupts
LCD_RS = CTRL_REG;      // Switch to instruction register
// Set LCD_DATA to high nibble of Display On/Off Control
LCD_DATA = (LCD_DATA&0x0F)|0x40;
LCD_E = 1; LCD_E = 0;   // Write data to display
// Set LCD_DATA to lower nibble of Display On/Off Control
LCD_DATA = (LCD_DATA&0x0F)|0x00;
LCD_E = 1; LCD_E = 0;   // Write data to display
EA = 1;                 // Enable all interrupts

delay(4);               // Wait 400s

// Write the characterpattern in sequence

LCD_RS = DATA_REG;              // Switch to data register

for ( count = 0; count < 64; count++)
  {
  EA = 0;                         // Disable all interrupts
  // Set LCD_DATA to high nibble of value
  LCD_DATA = (LCD_DATA&0x0F)|((pattern[count])&0xF0);
  LCD_E = 1; LCD_E = 0;           // Write data to display
  // Set LCD_DATA to lower nibble of value
  LCD_DATA = (LCD_DATA&0x0F)|(((pattern[count])<<4)&0xF0);
  LCD_E = 1; LCD_E = 0;           // Write data to display
  EA = 1;                         // Enable all interrupts
  delay(4);                       // Wait 400s
  }

return;
}


/********************************************************************/
/* Function    : LCD_control(dsp,blink,cursor)                  SUB */
/*------------------------------------------------------------------*/
/* Description : Routine controls the screen                        */
/*------------------------------------------------------------------*/
/* Author      : Dipl.-Ing. (FH) Thorsten Godau                     */
/*------------------------------------------------------------------*/
/* Input       : unsigned char dsp    = ON     -> Display on        */
/*                                      OFF    -> Display off       */
/*               unsigned char blink  = BLINK  -> Cursor blinks     */
/*                                      NOBLINK-> Cursor not blinks */
/*               unsigned char cursor = SHOW   -> Cursor visible    */
/*                                      HIDE   -> Cursor invisible  */
/*------------------------------------------------------------------*/
/* Returnvalue : none                                               */
/*------------------------------------------------------------------*/
/* History     : 06/99    V1.0 Basic routine                        */
/*               09/99    V1.1 Calculation of control modified      */
/*                             (because of compiler optim. bug)     */
/*                                                                  */
/********************************************************************/
void LCD_control(unsigned char dsp,unsigned char blink,unsigned char cursor)
{
unsigned char control;  // variable to generate instruction byte

control = (0x08 + blink + cursor + dsp); // Cursor control

EA = 0;                 // Disable all interrupts
LCD_RS = CTRL_REG;      // Switch to instruction register
// Set LCD_DATA to high nibble of Display On/Off Control
LCD_DATA = (LCD_DATA&0x0F)|0x00;
LCD_E = 1; LCD_E = 0;   // Write data to display
// Set LCD_DATA to lower nibble of Display On/Off Control
LCD_DATA = (LCD_DATA&0x0F)|((control<<4)&0xF0);
LCD_E = 1; LCD_E = 0;   // Write data to display
EA = 1;                 // Enable all interrupts

delay(4);               // Wait 400s

return;
}


/********************************************************************/
/* Function    : LCD_gotoxy(x,y)                                SUB */
/*------------------------------------------------------------------*/
/* Description : Sets the write position of the LCD display         */
/*                                                                  */
/*                 (1,1)         (16,1)                             */
/*                   |              |                               */
/*                   ################   -> line 1                   */
/*                   ################   -> line 2                   */
/*                   ################   -> line 3 (optional)        */
/*                   ################   -> line 4 (optional)        */
/*                   |              |                               */
/*                 (1,4)         (16,4)                             */
/*------------------------------------------------------------------*/
/* Author      : Dipl.-Ing. (FH) Thorsten Godau                     */
/*------------------------------------------------------------------*/
/* Input       : unsigned char x    -> x position (horizontal)      */
/*               unsigned char y    -> y position (vertical)        */
/*------------------------------------------------------------------*/
/* Returnvalue : none                                               */
/*------------------------------------------------------------------*/
/* History     : 06/99    V1.0 Basic routine                        */
/*               01/02    V1.1 Array LOCATION expanded to 4-line    */
/*                                                                  */
/********************************************************************/
void LCD_gotoxy(unsigned char x,unsigned char y)
{

EA = 0;                         // Disable all interrupts
LCD_RS = CTRL_REG;              // Switch to instruction register
// Set LCD_DATA to high nibble of position table value
LCD_DATA = (LCD_DATA&0x0F)|((LOCATION[y-1][x-1])&0xF0);
LCD_E = 1; LCD_E = 0;           // Write data to display
// Set LCD_DATA to lower nibble of position table value
LCD_DATA = (LCD_DATA&0x0F)|(((LOCATION[y-1][x-1])<<4)&0xF0);
LCD_E = 1; LCD_E = 0;           // Write data to display
EA = 1;                         // Enable all interrupts

delay(4);                       // Wait 400s

return;
}


/********************************************************************/
/* Function    : LCD_clrscr()                                   SUB */
/*------------------------------------------------------------------*/
/* Description : Clears LCD display, and sets position to (1,1)     */
/*------------------------------------------------------------------*/
/* Author      : Dipl.-Ing. (FH) Thorsten Godau                     */
/*------------------------------------------------------------------*/
/* Input       : none                                               */
/*------------------------------------------------------------------*/
/* Returnvalue : none                                               */
/*------------------------------------------------------------------*/
/* History     : 06/99    V1.0 Basic routine                        */
/*               09/99    V1.1 Timing correction                    */
/*                                                                  */
/********************************************************************/
void LCD_clrscr(void)
{

EA = 0;                         // Disable all interrupts
LCD_RS = CTRL_REG;              // Switch to instruction register
// Set LCD_DATA to high nibble of Clear Screen
LCD_DATA = (LCD_DATA&0x0F)|0x00;
LCD_E = 1; LCD_E = 0;           // Write data to display
// Set LCD_DATA to lower nibble of Clear Screen
LCD_DATA = (LCD_DATA&0x0F)|0x10;
LCD_E = 1; LCD_E = 0;           // Write data to display
EA = 1;                         // Enable all interrupts

delay(41);                      // Wait 4.1ms

return;
}


/********************************************************************/
/* Function    : LCD_putchar(value)                             SUB */
/*------------------------------------------------------------------*/
/* Description : Writes the character value to the display          */
/*------------------------------------------------------------------*/
/* Author      : Dipl.-Ing. (FH) Thorsten Godau                     */
/*------------------------------------------------------------------*/
/* Input       : unsigned char value    -> character to be written  */
/*------------------------------------------------------------------*/
/* Returnvalue : none                                               */
/*------------------------------------------------------------------*/
/* History     : 06/99    V1.0 Basic routine                        */
/*                                                                  */
/********************************************************************/
void LCD_putchar(unsigned char value)
{

EA = 0;                         // Disable all interrupts
LCD_RS = DATA_REG;              // Switch to data register
// Set LCD_DATA to high nibble of value
LCD_DATA = (LCD_DATA&0x0F)|(value&0xF0);
LCD_E = 1; LCD_E = 0;           // Write data to display
// Set LCD_DATA to lower nibble of value
LCD_DATA = (LCD_DATA&0x0F)|((value<<4)&0xF0);
LCD_E = 1; LCD_E = 0;           // Write data to display
EA = 1;                         // Enable all interrupts

delay(4);                       // Wait 400s

return;
}


/********************************************************************/
/* Function    : LCD_printxy(x,y,*text)                         SUB */
/*------------------------------------------------------------------*/
/* Description : Prints text to position x/y of the display         */
/*------------------------------------------------------------------*/
/* Author      : Dipl.-Ing. (FH) Thorsten Godau                     */
/*------------------------------------------------------------------*/
/* Input       : unsigned char x     -> x position of the display   */
/*               unsigned char y     -> y position of the display   */
/*               unsigned char *text -> pointer to text             */
/*------------------------------------------------------------------*/
/* Returnvalue : none                                               */
/*------------------------------------------------------------------*/
/* History     : 06/99    V1.0 Basic routine                        */
/*                                                                  */
/********************************************************************/
void LCD_printxy(unsigned char x,unsigned char y,unsigned char *text)
{

LCD_gotoxy(x,y);        // Set cursor position

while( *text )          // While not end of text
  {
  LCD_putchar(*text++); // Write character and increment position
  }

return;
}


/********************************************************************/
/* Function    : delay(delaytime)                               SUB */
/*------------------------------------------------------------------*/
/* Description : Routine waits (dependening on Timer 1) for about   */
/*               delaytime * 100s. Best for delaytime > 10         */
/*------------------------------------------------------------------*/
/* Author      : Dipl.-Ing. (FH) Thorsten Godau                     */
/*------------------------------------------------------------------*/
/* Input       : unsigned int time_100s                            */
/*------------------------------------------------------------------*/
/* Returnvalue : none                                               */
/*------------------------------------------------------------------*/
/* History     : 06/99    V1.0 Basic routine                        */
/*               09/99    V1.1 nop inserted in delay-loop           */
/*               11/99    V1.2 Removed nop for better accuracy      */
/*                                                                  */
/********************************************************************/
void delay(unsigned int time_100us)
{

if ( time_100us == 0 ) return;   // Return if no delaytime is assigned

DELAY = time_100us;              // Set global delaytime

while ( DELAY > 0 );

return;
}


/********************************************************************/
/* Function    : putchar()                                      SUB */
/*------------------------------------------------------------------*/
/* Description : Routine sends the character value via the UART     */
/*               (can also be used by printf)                       */
/*------------------------------------------------------------------*/
/* Author      : Dipl.-Ing. (FH) Thorsten Godau                     */
/*------------------------------------------------------------------*/
/* Input       : unsigned char value    0x00...0xFF                 */
/*------------------------------------------------------------------*/
/* Returnvalue : none                                               */
/*------------------------------------------------------------------*/
/* History     : 06/99    V1.0 Basic routine                        */
/*               09/99    V1.1 nop inserted in wait-loop            */
/*                                                                  */
/********************************************************************/
void putchar(unsigned char value)
{

while ( TXCNT >= BSIZE )  /* Wait for room in buffer */
  {
  _asm
    nop
  _endasm;
  }

ES = 0;

if ( BUSY )
  {
  TXBuffer[( TXPOS + TXCNT++ ) % BSIZE] = value;
  }
else
  {
  SBUF = value;
  BUSY = 1;
  }

ES = 1;

return;
}


/********************************************************************/
/* Function    : UART_puts(*text)                               SUB */
/*------------------------------------------------------------------*/
/* Description : Prints text to the UART                            */
/*------------------------------------------------------------------*/
/* Author      : Dipl.-Ing. (FH) Thorsten Godau                     */
/*------------------------------------------------------------------*/
/* Input       : unsigned char *text -> pointer to text             */
/*------------------------------------------------------------------*/
/* Returnvalue : none                                               */
/*------------------------------------------------------------------*/
/* History     : 06/99    V1.0 Basic routine                        */
/*                                                                  */
/********************************************************************/
void UART_puts(unsigned char *text)
{

while( *text )          // While not end of text
  {
  putchar(*text++);     // Write character and increment position
  }

return;
}


/********************************************************************/
/* Function    : RXChar()                                       SUB */
/*------------------------------------------------------------------*/
/* Description : This function returns a character from the         */
/*               receivebuffer if available.                        */
/*------------------------------------------------------------------*/
/* Author      : Dipl.-Ing. (FH) Thorsten Godau                     */
/*------------------------------------------------------------------*/
/* Input       : none                                               */
/*------------------------------------------------------------------*/
/* Returnvalue : unsigned int (range 0x0000..0x0100)                */
/*                                                                  */
/*                     0x0100   (EOF) no character in rxbuffer      */
/*                     0x0000..0x00FF character from UART           */
/*------------------------------------------------------------------*/
/* History     : 06/99    V1.0 Basic routine                        */
/*                                                                  */
/********************************************************************/
unsigned int RXChar(void)
{
unsigned char value;        // Declare value variable

if ( RXE != RXB )           // Character available
  {
  value = RXBuffer[RXE];

  if ( ++RXE >= BSIZE )     // Check end of buffer
    RXE = 0;
  }
else
  {
  return(NCA);              // No character available, return NCA
  }

return((unsigned int)value);
}


/********************************************************************/
/* Function    : i2c_delay()                                    SUB */
/*------------------------------------------------------------------*/
/* Description : This subroutine waits for 5s. Needed for I2C.     */
/*               (Low-level I2C routine)                            */
/*------------------------------------------------------------------*/
/* Author      : Dipl.-Ing. (FH) Thorsten Godau                     */
/*               (Based upon U24C08 EEPROM programming with C51 by  */
/*               Dimitry S. Obukhov dso@usa.net in 1996/97)         */
/*------------------------------------------------------------------*/
/* Input       : none                                               */
/*------------------------------------------------------------------*/
/* Returnvalue : none                                               */
/*------------------------------------------------------------------*/
/* History     : 11/99    V1.0 Basic routine                        */
/*                                                                  */
/********************************************************************/
void i2c_delay(void)
{

EA = 0;    // Disable all interrupts for a accurate timing

// Valid for fq = 24MHz -> Tcycle = 500ns -> 10 x nop
_asm
  nop
  nop
  nop
  nop
  nop
  nop
  nop
  nop
  nop
  nop
_endasm;

EA = 1;    // Enable all interrupts

return;
}


/********************************************************************/
/* Function    : i2c_start()                                    SUB */
/*------------------------------------------------------------------*/
/* Description : This subroutine generates a I2C START condition.   */
/*               (Low-level I2C routine)                            */
/*------------------------------------------------------------------*/
/* Author      : Dipl.-Ing. (FH) Thorsten Godau                     */
/*               (Based upon U24C08 EEPROM programming with C51 by  */
/*               Dimitry S. Obukhov dso@usa.net in 1996/97)         */
/*------------------------------------------------------------------*/
/* Input       : none                                               */
/*------------------------------------------------------------------*/
/* Returnvalue : none                                               */
/*------------------------------------------------------------------*/
/* History     : 11/99    V1.0 Basic routine                        */
/*                                                                  */
/********************************************************************/
void i2c_start(void)
{

EA = 0;        // Disable all interrupts for a accurate timing

// Force defined state of SCL and SDA
SCL = 1;       // Release SCL
SDA = 1;       // Release SDA
i2c_delay();   // Force 5s-delay

// Generate START condition
SDA = 0;
i2c_delay();   // Force 5s-delay
SCL = 0;

EA = 1;        // Enable all interrupts

// I2C Busy

return;
}


/********************************************************************/
/* Function    : i2c_stop()                                     SUB */
/*------------------------------------------------------------------*/
/* Description : This subroutine generates a I2C STOP condition.    */
/*               (Low-level I2C routine)                            */
/*------------------------------------------------------------------*/
/* Author      : Dipl.-Ing. (FH) Thorsten Godau                     */
/*               (Based upon U24C08 EEPROM programming with C51 by  */
/*               Dimitry S. Obukhov dso@usa.net in 1996/97)         */
/*------------------------------------------------------------------*/
/* Input       : none                                               */
/*------------------------------------------------------------------*/
/* Returnvalue : none                                               */
/*------------------------------------------------------------------*/
/* History     : 11/99    V1.0 Basic routine                        */
/*                                                                  */
/********************************************************************/
void i2c_stop(void)
{

EA = 0;        // Disable all interrupts for a accurate timing

// Generate STOP condition
SDA = 0;
i2c_delay();   // Force 5s-delay
SCL = 1;       // Release SCL
i2c_delay();   // Force 5s-delay
SDA = 1;       // Release SDA

EA = 1;        // Enable all interrupts

// I2C Idle

return;
}


/********************************************************************/
/* Function    : i2c_ack()                                      SUB */
/*------------------------------------------------------------------*/
/* Description : This subroutine generates a I2C MASTER ACK.        */
/*               MASTER clock synchronisation (needed to handle     */
/*               SLAVE clock streching) is supported.               */
/*               (Low-level I2C routine)                            */
/*------------------------------------------------------------------*/
/* Author      : Dipl.-Ing. (FH) Thorsten Godau                     */
/*               (Based upon U24C08 EEPROM programming with C51 by  */
/*               Dimitry S. Obukhov dso@usa.net in 1996/97)         */
/*------------------------------------------------------------------*/
/* Input       : none                                               */
/*------------------------------------------------------------------*/
/* Returnvalue : none                                               */
/*------------------------------------------------------------------*/
/* History     : 11/99    V1.0 Basic routine                        */
/*                                                                  */
/********************************************************************/
void i2c_ack(void)
{

EA = 0;            // Disable all interrupts for a accurate timing

// Generate ACK
SDA = 0;
SCL = 1;           // Release SCL
while(SCL == 0);   // Synchronize clock
i2c_delay();       // Force 5s-delay
SCL = 0;           // Force a clock cycle

EA = 1;            // Enable all interrupts

return;
}


/********************************************************************/
/* Function    : i2c_nack()                                     SUB */
/*------------------------------------------------------------------*/
/* Description : This subroutine generates a I2C MASTER NACK.       */
/*               MASTER clock synchronisation (needed to handle     */
/*               SLAVE clock streching) is supported.               */
/*               (Low-level I2C routine)                            */
/*------------------------------------------------------------------*/
/* Author      : Dipl.-Ing. (FH) Thorsten Godau                     */
/*               (Based upon U24C08 EEPROM programming with C51 by  */
/*               Dimitry S. Obukhov dso@usa.net in 1996/97)         */
/*------------------------------------------------------------------*/
/* Input       : none                                               */
/*------------------------------------------------------------------*/
/* Returnvalue : none                                               */
/*------------------------------------------------------------------*/
/* History     : 11/99    V1.0 Basic routine                        */
/*                                                                  */
/********************************************************************/
void i2c_nack(void)
{

EA = 0;            // Disable all interrupts for a accurate timing

// Generate NACK
SDA = 1;           // Release SDA
SCL = 1;           // Release SCL
while(SCL == 0);   // Synchronize clock
i2c_delay();       // Force 5s-delay
SCL = 0;           // Force a clock cycle

EA = 1;            // Enable all interrupts

return;
}


/********************************************************************/
/* Function    : i2c_check()                                    SUB */
/*------------------------------------------------------------------*/
/* Description : This subroutine checkes an I2C SLAVE ACK.          */
/*               MASTER clock synchronisation (needed to handle     */
/*               SLAVE clock streching) is supported.               */
/*               (Low-level I2C routine)                            */
/*------------------------------------------------------------------*/
/* Author      : Dipl.-Ing. (FH) Thorsten Godau                     */
/*               (Based upon U24C08 EEPROM programming with C51 by  */
/*               Dimitry S. Obukhov dso@usa.net in 1996/97)         */
/*------------------------------------------------------------------*/
/* Input       : none                                               */
/*------------------------------------------------------------------*/
/* Returnvalue : unsigned char -> 0 : acknowledge from Slave        */
/*                                1 : no acknowledge from Slave     */
/*------------------------------------------------------------------*/
/* History     : 11/99    V1.0 Basic routine                        */
/*                                                                  */
/********************************************************************/
unsigned char i2c_check(void)
{

EA = 0;            // Disable all interrupts for a accurate timing

SDA = 1;           // Release SDA
SCL = 1;           // Release SCL
while(SCL == 0);   // Synchronize clock
i2c_delay();       // Force 5s-delay
if (SDA)           // SDA is high
  {
  SCL = 0;         // Force a clock cycle
  EA = 1;          // Enable all interrupts
  return(1);       // No acknowledge from Slave
  }
SCL = 0;           // Force a clock cycle

EA = 1;            // Enable all interrupts

return(0);         // Acknowledge from Slave
}


/********************************************************************/
/* Function    : i2c_write(value)                               SUB */
/*------------------------------------------------------------------*/
/* Description : This subroutine writes a character to I2C.         */
/*               MASTER clock synchronisation (needed to handle     */
/*               SLAVE clock streching) is supported.               */
/*               (Low-level I2C routine)                            */
/*------------------------------------------------------------------*/
/* Author      : Dipl.-Ing. (FH) Thorsten Godau                     */
/*               (Based upon U24C08 EEPROM programming with C51 by  */
/*               Dimitry S. Obukhov dso@usa.net in 1996/97)         */
/*------------------------------------------------------------------*/
/* Input       : unsigned char value -> data to I2C device          */
/*------------------------------------------------------------------*/
/* Returnvalue : unsigned char       -> 0 : Acknowledge received    */
/*                                      1 : No acknowledge received */
/*------------------------------------------------------------------*/
/* History     : 11/99    V1.0 Basic routine                        */
/*                                                                  */
/********************************************************************/
unsigned char i2c_write(unsigned char value)
{
idata unsigned char counter = 0;  // Bitcounter

EA = 0;                           // Disable all interrupts for a accurate timing

for ( counter = 0; counter < 8; counter++ )
  {
  SDA = (bit)((value&0x80) >> 7); // Set SDA to value of MSB
  SCL = 1;                        // Release SCL
  while(SCL == 0);                // Synchronize clock
  i2c_delay();                    // Force 5s-delay
  SCL = 0;                        // Force a clock cycle
  value <<= 1;                    // Prepare next bit for transmission
  }

EA = 1;                           // Enable all interrupts

// Generate a 9. clock cycle and check ACK from SLAVE
// Return the result
return(i2c_check());
}


/********************************************************************/
/* Function    : i2c_read()                                     SUB */
/*------------------------------------------------------------------*/
/* Description : This subroutine reads a character from I2C.        */
/*               MASTER clock synchronisation (needed to handle     */
/*               SLAVE clock streching) is supported.               */
/*               (Low-level I2C routine)                            */
/*------------------------------------------------------------------*/
/* Author      : Dipl.-Ing. (FH) Thorsten Godau                     */
/*               (Based upon U24C08 EEPROM programming with C51 by  */
/*               Dimitry S. Obukhov dso@usa.net in 1996/97)         */
/*------------------------------------------------------------------*/
/* Input       : none                                               */
/*------------------------------------------------------------------*/
/* Returnvalue : unsigned char -> data from I2C device              */
/*------------------------------------------------------------------*/
/* History     : 11/99    V1.0 Basic routine                        */
/*                                                                  */
/********************************************************************/
unsigned char i2c_read(void)
{
idata unsigned char result  = 0;  // Returnvalue with read I2C byte
idata unsigned char counter = 0;  // Bitcounter

EA = 0;              // Disable all interrupts for a accurate timing

for ( counter = 0; counter < 8; counter++ )
  {
  SDA = 1;           // Release SDA
  SCL = 1;           // Release SCL
  while(SCL == 0);   // Synchronize clock
  i2c_delay();       // Force 5s-delay

  result <<= 1;      // Shift left the result

  if (SDA)
    result |= 0x01;  // Set actual SDA state to LSB

  SCL = 0;           // Force a clock cycle
  i2c_delay();       // Force 5s-delay
  }

EA = 1;              // Enable all interrupts

// ACK or NACK from MASTER must be generated outside this routine

return(result);
}


/********************************************************************/
/* Function    : EEPROM_write(adress,content)                   SUB */
/*------------------------------------------------------------------*/
/* Description : This subroutine writes content to adress of EEPROM */
/*               For use with Xicor X24C02 (256 Byte x 8)           */
/*               (High-level I2C routine)                           */
/*------------------------------------------------------------------*/
/* Author      : Dipl.-Ing. (FH) Thorsten Godau                     */
/*------------------------------------------------------------------*/
/* Input       : unsigned char adress  -> mem adress (0x00..0xFF)   */
/*               unsigned char content -> mem data   (0x00..0xFF)   */
/*------------------------------------------------------------------*/
/* Returnvalue : none                                               */
/*------------------------------------------------------------------*/
/* History     : 11/99    V1.0 Basic routine                        */
/*                                                                  */
/********************************************************************/
void EEPROM_write(unsigned char adress,unsigned char content)
{

i2c_start();   // Generate START condition

// Send SLAVE adress with write request
if ( i2c_write(EEPROM_ID|WRITE) )
  {
  errormsg('1');
  return;
  }

// Send memory adress
if ( i2c_write(adress) )
  {
  errormsg('2');
  return;
  }

// Send content to memory adress
if ( i2c_write(content) )
  {
  errormsg('3');
  return;
  }

i2c_stop();    // Generate STOP condition

delay(100);    // Secure wait for the EEPROMs write cycle

return;
}


/********************************************************************/
/* Function    : EEPROM_read()                                  SUB */
/*------------------------------------------------------------------*/
/* Description : This subroutine reads content from EEPROM adress.  */
/*               The random read operation is used. Hence after     */
/*               receiving the content, a STOP condition without a  */
/*               MASTER ACK before is performed !                   */
/*               For use with Xicor X24C02 (256 Byte x 8)           */
/*               (High-level I2C routine)                           */
/*------------------------------------------------------------------*/
/* Author      : Dipl.-Ing. (FH) Thorsten Godau                     */
/*------------------------------------------------------------------*/
/* Input       : unsigned char adress  -> mem adress (0x00..0xFF)   */
/*------------------------------------------------------------------*/
/* Returnvalue : unsigned char         -> mem data   (0x00..0xFF)   */
/*------------------------------------------------------------------*/
/* History     : 11/99    V1.0 Basic routine                        */
/*                                                                  */
/********************************************************************/
unsigned char EEPROM_read(unsigned char adress)
{
idata unsigned char result = 0;

i2c_start();   // Generate START condition

// Send SLAVE adress with dummy write request
if ( i2c_write(EEPROM_ID|WRITE) )
  {
  errormsg('4');
  return(0);
  }

// Send memory adress
if ( i2c_write(adress) )
  {
  errormsg('5');
  return(0);
  }

i2c_start();   // Generate START condition

// Send SLAVE adress with read request
if ( i2c_write(EEPROM_ID|READ) )
  {
  errormsg('6');
  return(0);
  }

result = i2c_read();   // Read memory content
// Dont perform a MASTER ACK !
i2c_stop();            // Generate STOP condition

return(result);
}


/********************************************************************/
/* Function    : DPOT_write(potnum,content)                     SUB */
/*------------------------------------------------------------------*/
/* Description : This subroutine writes content to DPOT #potnum.    */
/*               For use with Dallas DS1803 Dualpot                 */
/*               (High-level I2C routine)                           */
/*------------------------------------------------------------------*/
/* Author      : Dipl.-Ing. (FH) Thorsten Godau                     */
/*------------------------------------------------------------------*/
/* Input       : bit           potnum  -> 0 : Pot-0                 */
/*                                        1 : Pot-1                 */
/*               unsigned char content -> wiper pos. (0x00..0xFF)   */
/*------------------------------------------------------------------*/
/* Returnvalue : none                                               */
/*------------------------------------------------------------------*/
/* History     : 11/99    V1.0 Basic routine                        */
/*                                                                  */
/********************************************************************/
void DPOT_write(bit potnum, unsigned char content)
{

i2c_start();   // Generate START condition

// Send SLAVE adress with write request
if ( i2c_write(DPOT_ID|WRITE) )
  {
  errormsg('7');
  return;
  }

// Send command byte
if ( i2c_write(0xA9+(unsigned char)potnum) )
  {
  errormsg('8');
  return;
  }

// Send wiper position
if ( i2c_write(content) )
  {
  errormsg('9');
  return;
  }

i2c_stop();    // Generate STOP condition

delay(10);     // Secure wait for the DPOTs write cycle

return;
}


/********************************************************************/
/* Function    : ADC_read()                                     SUB */
/*------------------------------------------------------------------*/
/* Description : This subroutine reads the channel adchannel of the */
/*               AD/DA IC.                                          */
/*               For use with Philips PCF8591 ADC/DAC               */
/*               (High-level I2C routine)                           */
/*------------------------------------------------------------------*/
/* Author      : Dipl.-Ing. (FH) Thorsten Godau                     */
/*------------------------------------------------------------------*/
/* Input       : unsigned char adchannel -> 0 : A/D channel 0       */
/*                                          1 : A/D channel 1       */
/*                                          2 : A/D channel 2       */
/*                                          3 : A/D channel 3       */
/*------------------------------------------------------------------*/
/* Returnvalue : unsigned char           -> ADC result (0x00..0xFF) */
/*------------------------------------------------------------------*/
/* History     : 11/99    V1.0 Basic routine                        */
/*                                                                  */
/********************************************************************/
unsigned char ADC_read(unsigned char adchannel)
{
idata unsigned char result = 0;

i2c_start();   // Generate START condition

// Send SLAVE adress with write request
if ( i2c_write(ADDA_ID|WRITE) )
  {
  errormsg('A');
  return(0);
  }

// Send control byte :
// DAC off, all ADCs single ended, no autoinc, channel
if ( i2c_write(ADCTRL|adchannel|DACEN) )
  {
  errormsg('B');
  return(0);
  }

i2c_start();   // Generate START condition

// Send SLAVE adress with read request
if ( i2c_write(ADDA_ID|READ) )
  {
  errormsg('C');
  return(0);
  }

// Here starts the AD-conversion

i2c_read();    // Perform a dummy read of a byte
i2c_ack();     // Perform a MASTER ACK

// Here comes the requested byte !

result = i2c_read();
i2c_nack();    // Perform a MASTER NACK

i2c_stop();    // Generate STOP condition

return(result);
}


/********************************************************************/
/* Function    : DAC_write()                                    SUB */
/*------------------------------------------------------------------*/
/* Description : This subroutine writes the DAC of the AD/DA IC.    */
/*               For use with Philips PCF8591 ADC/DAC               */
/*               (High-level I2C routine)                           */
/*------------------------------------------------------------------*/
/* Author      : Dipl.-Ing. (FH) Thorsten Godau                     */
/*------------------------------------------------------------------*/
/* Input       : unsigned char content -> DAC-value (0x00..0xFF)    */
/*------------------------------------------------------------------*/
/* Returnvalue : none                                               */
/*------------------------------------------------------------------*/
/* History     : 10/01    V1.0 Basic routine                        */
/*                                                                  */
/********************************************************************/
void DAC_write(unsigned char content)
{

i2c_start();   // Generate START condition

// Send SLAVE adress with write request
if ( i2c_write(ADDA_ID|WRITE) )
  {
  errormsg('D');
  return;
  }

// Send control byte :
// DAC on, all ADCs single ended, no autoinc
if ( i2c_write(ADCTRL|DACEN) )
  {
  errormsg('E');
  return;
  }

// Send content byte (dummy write)
i2c_write(content);
i2c_write(content);

// Here starts the DA-conversion

// Send content byte again to perform the DAC
if ( i2c_write(content) )
  {
  errormsg('F');
  return;
  }

i2c_stop();    // Generate STOP condition

return;
}


/********************************************************************/
/* Function    : gettime()                                      SUB */
/*------------------------------------------------------------------*/
/* Description : This subroutine gets the time of the RTC.          */
/*               The result is stored in struct TM.                 */
/*               For use with Philips PCF8583 RTC/RAM               */
/*               (High-level I2C routine)                           */
/*------------------------------------------------------------------*/
/* Author      : Dipl.-Ing. (FH) Thorsten Godau                     */
/*------------------------------------------------------------------*/
/* Input       : Pointer to struct TM                               */
/*------------------------------------------------------------------*/
/* Returnvalue : none                                               */
/*------------------------------------------------------------------*/
/* History     : 10/01    V1.0 Basic routine                        */
/*                                                                  */
/********************************************************************/
void gettime(struct TIME_T *time)
{
volatile unsigned char result;

i2c_start();   // Generate START condition

// Send SLAVE adress
if ( i2c_write(RTC_ID|WRITE) )
  {
  errormsg('G');
  return;
  }

// Send memory adress (register seconds)
if ( i2c_write(0x02) )
  {
  errormsg('H');
  return;
  }

i2c_start();   // Generate START condition

// Send SLAVE adress with read request
if ( i2c_write(RTC_ID|READ) )
  {
  errormsg('I');
  return;
  }

// Read seconds
result = i2c_read(); i2c_ack();
time->tm_sec  = ((result>>4)*10) + (result&0x0F);

// Read minutes
result = i2c_read(); i2c_ack();
time->tm_min  = ((result>>4)*10) + (result&0x0F);

// Read hours
result = i2c_read(); i2c_ack();
time->tm_hour = (((result&0x30)>>4)*10) + (result&0x0F);

// Read year and date
result = i2c_read(); i2c_ack();
time->tm_year = ((result&0xC0)>>6);
time->tm_dom  = (((result&0x30)>>4)*10) + (result&0x0F);

// Read weekday and month
result = i2c_read(); i2c_nack();
                     i2c_stop();            // Generate STOP condition
time->tm_dow  = ((result&0xE0)>>5);
time->tm_mon  = (((result&0x10)>>4)*10) + (result&0x0F);

return;  
}


/********************************************************************/
/* Function    : settime()                                      SUB */
/*------------------------------------------------------------------*/
/* Description : This subroutine sets the time of the RTC.          */
/*               Struct TM has to be filled before.                 */
/*               For use with Philips PCF8583 RTC/RAM               */
/*               (High-level I2C routine)                           */
/*------------------------------------------------------------------*/
/* Author      : Dipl.-Ing. (FH) Thorsten Godau                     */
/*------------------------------------------------------------------*/
/* Input       : Pointer to struct TM                               */
/*------------------------------------------------------------------*/
/* Returnvalue : none                                               */
/*------------------------------------------------------------------*/
/* History     : 11/01    V1.0 Basic routine                        */
/*                                                                  */
/********************************************************************/
void settime(struct TIME_T *time)
{
volatile unsigned char content;


i2c_start();   // Generate START condition

// Send SLAVE adress
if ( i2c_write(RTC_ID|WRITE) )
  {
  errormsg('G');
  return;
  }

// Send memory adress (register seconds)
if ( i2c_write(0x02) )
  {
  errormsg('H');
  return;
  }

// Send seconds
content = 0;
content = ( (((time->tm_sec)/10)<<4) | ((time->tm_sec)%10) );
if ( i2c_write(content) )
  {
  errormsg('J');
  return;
  }

// Send minutes
content = 0;
content = ( (((time->tm_min)/10)<<4) | ((time->tm_min)%10) );
if ( i2c_write(content) )
  {
  errormsg('J');
  return;
  }

// Send hours
if ( (time->tm_hour) > 12 )
  content = 0x40;
else
  content = 0x00;

content |= ( (((time->tm_hour)/10)<<4) | ((time->tm_hour)%10) );
if ( i2c_write(content) )
  {
  errormsg('J');
  return;
  }
  
// Send year and date
content  = 0;
content  = ( (time->tm_year)<<6 );
content |= ( (((time->tm_dom)/10)<<4) | ((time->tm_dom)%10) );
if ( i2c_write(content) )
  {
  errormsg('J');
  return;
  }

// Send weekday and month
content  = 0;
content  = ( (time->tm_dow)<<5 );
content |= ( (((time->tm_mon)/10)<<4) | ((time->tm_mon)%10) );
if ( i2c_write(content) )
  {
  errormsg('J');
  return;
  }

i2c_stop();    // Generate STOP condition

return;
}


/********************************************************************/
/* Function    : XIO_write()                                    SUB */
/*------------------------------------------------------------------*/
/* Description : This subroutine writes the output of the XIO IC.   */
/*               For use with Philips PCF8574 PIO                   */
/*               (High-level I2C routine)                           */
/*------------------------------------------------------------------*/
/* Author      : Dipl.-Ing. (FH) Thorsten Godau                     */
/*------------------------------------------------------------------*/
/* Input       : unsigned char content -> PIO-value (0x00..0xFF)    */
/*------------------------------------------------------------------*/
/* Returnvalue : none                                               */
/*------------------------------------------------------------------*/
/* History     : 11/01    V1.0 Basic routine                        */
/*                                                                  */
/********************************************************************/
void XIO_write(unsigned char content)
{

i2c_start();   // Generate START condition

// Send SLAVE adress with write request
if ( i2c_write(XIO_ID|WRITE) )
  {
  errormsg('K');
  return;
  }

// Send content byte
if ( i2c_write(content) )
  {
  errormsg('L');
  return;
  }

i2c_stop();    // Generate STOP condition

return;
}


/********************************************************************/
/* Function    : XIO_read()                                     SUB */
/*------------------------------------------------------------------*/
/* Description : This subroutine reads the input of the XIO IC.     */
/*               For use with Philips PCF8574 PIO                   */
/*               (High-level I2C routine)                           */
/*------------------------------------------------------------------*/
/* Author      : Dipl.-Ing. (FH) Thorsten Godau                     */
/*------------------------------------------------------------------*/
/* Input       : none                                               */
/*------------------------------------------------------------------*/
/* Returnvalue : unsigned char -> XIO-value (0x00..0xFF)            */
/*------------------------------------------------------------------*/
/* History     : 11/01    V1.0 Basic routine                        */
/*                                                                  */
/********************************************************************/
unsigned char XIO_read(void)
{
unsigned char result;

i2c_start();   // Generate START condition

// Send SLAVE adress with write request
if ( i2c_write(XIO_ID|WRITE) )
  {
  errormsg('K');
  return(0);
  }

// Set all pins high to be used as inputs
if ( i2c_write(0xFF) )
  {
  errormsg('L');
  return(0);
  }

i2c_stop();    // Generate STOP condition

i2c_start();   // Generate START condition

// Send SLAVE adress with read request
if ( i2c_write(XIO_ID|READ) )
  {
  errormsg('M');
  return(0);
  }

// Read port
result = i2c_read();

i2c_stop();    // Generate STOP condition

return(result);
}


/********************************************************************/
/* Function    : errormsg(ascii)                                SUB */
/*------------------------------------------------------------------*/
/* Description : This subroutine generates a error message via UART */
/*------------------------------------------------------------------*/
/* Author      : Dipl.-Ing. (FH) Thorsten Godau                     */
/*------------------------------------------------------------------*/
/* Input       : unsigned int ascii -> Character sent via UART      */
/*------------------------------------------------------------------*/
/* Returnvalue : none                                               */
/*------------------------------------------------------------------*/
/* History     : 11/99    V1.0 Basic routine                        */
/*                                                                  */
/********************************************************************/
void errormsg(unsigned char ascii)
{
putchar('');
putchar(ascii);
UART_puts("-error occured !");
putchar(0x0A);
putchar(0x0D);
putchar(0x07);

return;
}


/********************************************************************/
/* Function    : softbacklight()                                SUB */
/*------------------------------------------------------------------*/
/* Description : This subroutine turns the LCD backlight in smooth  */
/*               steps on or off                                    */
/*------------------------------------------------------------------*/
/* Author      : Dipl.-Ing. (FH) Thorsten Godau                     */
/*------------------------------------------------------------------*/
/* Input       : bit direction -> 0 : turn backlight off            */
/*                                1 : turn backlight on             */
/*------------------------------------------------------------------*/
/* Returnvalue : none                                               */
/*------------------------------------------------------------------*/
/* History     : 11/01    V1.0 Basic routine                        */
/*                                                                  */
/********************************************************************/
void softbacklight(bit direction)
{
unsigned char period;
unsigned char on_time;
unsigned char step;

if ( direction == 0 )
  {
  LCD_BL = 0;  // Backlight on
  for (step = 255; step > 0; step--)
    {
    on_time = 255 - step;
    for (period = 255; period > 0; period--)
      {
      if (on_time <= period) LCD_BL = 0; // On
      else                   LCD_BL = 1; // Off
      } 
    } 
  }
else
  {
  LCD_BL = 1;  // Backlight off
  for (step = 255; step > 0; step--)
    {
    on_time = 255 - step;
    for (period = 255; period > 0; period--)
      {
      if (on_time <= period) LCD_BL = 1; // Off
      else                   LCD_BL = 0; // On
      } 
    } 
  }

return;
}


// MAIN =============================================================
void main(void)
{
volatile idata unsigned int rxbyte;   // Contains the byte received from UART

volatile unsigned char dummy;         // General purpose
volatile unsigned int  errors;        // Errorcounter

init_MCU();                           // Initialize MCU

putchar(0x0A); putchar(0x0D);                          // CR/LF via UART
UART_puts("# > uAnt51 1V10 < (C) 2001 DL9SEC");        // Message via UART
putchar(0x0A); putchar(0x0D);                          // CR/LF via UART
UART_puts("# MCU ready...");                           // Message via UART
putchar(0x0A);putchar(0x0D);                           // CR/LF via UART
putchar(0x07);                                         // Ctrl-G (Bell) via UART

LCD_init();                                            // Initialize the LCD display

LCD_printxy(1,1,"> uAnt51  1V10 <");                   // Message to LCD
LCD_printxy(1,2,"(C) 2001 DL9SEC ");                   // Message to LCD
LCD_control(ON,NOBLINK,HIDE);                          // Display on, cursor hidden and non-blinking
delay(50000);                                          // Wait for 2 sec

LCD_printxy(1,2,"Selftest BEDCB  ");                   // Message to LCD
LCD_gotoxy(10,2);

// Test of the beeper CW : OK --- -.-
UART_puts("# Testing beeper...");                      // Message via UART
putchar(0x0A); putchar(0x0D);                          // CR/LF via UART

BEEPER = 0; delay(6000); BEEPER = 1; delay(2000);
BEEPER = 0; delay(6000); BEEPER = 1; delay(2000);
BEEPER = 0; delay(6000); BEEPER = 1; delay(6000);
BEEPER = 0; delay(6000); BEEPER = 1; delay(2000);
BEEPER = 0; delay(2000); BEEPER = 1; delay(2000);
BEEPER = 0; delay(6000); BEEPER = 1;

UART_puts("# Test done...");                           // Message via UART
putchar(0x0A); putchar(0x0D);                          // CR/LF via UART
LCD_putchar(0xFF);                                     // Update progress bar
delay(30000);
// End of beeper-test


// Test of the EEPROM via I2C
UART_puts("# Testing EEPROM...");                      // Message via UART
putchar(0x0A); putchar(0x0D);                          // CR/LF via UART

for (dummy = 0; dummy < 255; dummy++)
  {
  EEPROM_write(dummy, dummy);
  }

errors = 0;

for (dummy = 0; dummy < 255; dummy++)
  {
  if (EEPROM_read(dummy) != dummy)
    errors++;
  }
  
if (errors == 0)
  {
  UART_puts("# EEPROM OK...");                         // Message via UART
  putchar(0x0A); putchar(0x0D);                        // CR/LF via UART
  LCD_putchar(0xFF);                                   // Update progress bar
  }
else
  {  
  UART_puts("# EEPROM NOT OK...");                     // Message via UART
  putchar(0x0A); putchar(0x0D);                        // CR/LF via UART
  LCD_putchar(0xDB);                                   // Update progress bar
  }

delay(30000);
// End of EEPROM-test

// Test of the DAC
UART_puts("# Testing DAC...");                         // Message via UART
putchar(0x0A); putchar(0x0D);                          // CR/LF via UART
  
DAC_write(0x80);  // Set DAC to 2.5V

delay(10000);
UART_puts("# Test done...");                           // Message via UART
putchar(0x0A); putchar(0x0D);                          // CR/LF via UART
LCD_putchar(0xFF);                                     // Update progress bar

delay(30000);
// End of DAC-test

// Test of the RTC
// Preset TM
/*
TM.tm_sec  = 0;
TM.tm_min  = 21;
TM.tm_hour = 19;
TM.tm_dom  = 13;
TM.tm_mon  = 6;
TM.tm_year = 2;
TM.tm_dow  = 4;

settime(&TM);  // Write TM to RTC
*/

LCD_putchar(0xFF);
// End of RTC-test

softbacklight(1);    // LCD backlight on
softbacklight(0);    // LCD backlight off
LCD_putchar(0xFF);   // Update progress bar

delay(30000);

dummy = 0;

LCD_printxy(1,2,">          Uhr <");                //Time to LCD

softbacklight(1);    // LCD backlight on

while(1) // Main loop
  {
  if ( ( (rxbyte = RXChar()) & 0x100 ) != NCA )     // Loop incoming bytes via UART
    {
    putchar((unsigned char)(rxbyte & 0x00FF));      // Echo incoming bytes
    }

  gettime(&TM);                          // Read the time to LCD and store it in TM
  if ( TM.tm_sec != dummy )              // If new second
    {
    // Test of the RTC
    LCD_gotoxy(3,2);
    // Display time at LCD
    LCD_putchar(((TM.tm_hour)/10)+48);
    LCD_putchar(((TM.tm_hour)%10)+48);
    LCD_putchar(':');
    LCD_putchar(((TM.tm_min)/10)+48);
    LCD_putchar(((TM.tm_min)%10)+48);
    LCD_putchar('.');
    LCD_putchar(((TM.tm_sec)/10)+48);
    LCD_putchar(((TM.tm_sec)%10)+48);
    dummy = TM.tm_sec;                   // Store actual second
    // End of RTC-test
    LED_BL ^= 1;                         // Toggle blue LED
    }
  
  }


}
// END MAIN =========================================================